
/****************************************************************************
* Kurt's Brat - Running on an Arduino...
*
****************************************************************************/
#define DEBUG
//#define DEBUG_VERBOSE
//#define DEBUG_TIMING

//--------------------------------------------------------------------------
// Included libraries
//--------------------------------------------------------------------------
#include <PS2X_lib.h>
#include <Servo.h>
#include <EEPROM.h>
#include <pins_arduino.h>
#include <Streaming.h>
#include <avr/pgmspace.h>


#include "MoveData.h"
#define MNUMSERVOS 7
#include "ServoGroupMove.h"
extern void MSound(uint8_t _pin, byte cNotes, ...);

//--------------------------------------------------------------------------
// Global classes and variables
//--------------------------------------------------------------------------

PS2X ps2x; // create PS2 Controller Class

unsigned long 	g_tcLastActive;		// timer of when we were last active
boolean 	g_fCommandActive;
unsigned long 	g_tcPauseStart;		// time when we started a pause
unsigned long 	g_tcPausePeriod;	// how much time to pause.

unsigned long	g_tcLastDebug;		
word            g_LHipServoMS = 1500;    // Where we are when a command ends... used to angle the hips
word            g_RHipServoMS = 1500;


#if 0// defined(UBRR1H) 
// ATMEGA or the like, lets assume Botboarduino Mega shield
// Define the pin numbers we are using
#define RAS_PIN      2
#define RKS_PIN      3
#define RHS_PIN      4
#define LAS_PIN      5
#define LKS_PIN      6
#define LHS_PIN      9
#define TRTS_PIN     8


#define IRPIN    6
#define SOUND_PIN  46                // On Botboarduino Mega shield
#define PS2_DAT    30
#define PS2_CMD    31
#define PS2_SEL    32
#define PS2_CLK    33

#define AVS_PIN    15
#define AVAL_MIN    230 // about 4.5v with our 10K/30K resistor voltage divider
#else
// Lets assume Botboarduino JR board...
#define RAS_PIN    3 //A0
#define RKS_PIN    4  //A1
#define RHS_PIN    10 //A2
#define LAS_PIN    2  //2
#define LKS_PIN    11  //3
#define LHS_PIN    12  //4
#define TRTS_PIN   13  //10

#define IRPIN    3

#define SOUND_PIN    5        // Botboarduino JR pin number
#define PS2_DAT      6        
#define PS2_CMD      7
#define PS2_SEL      8
#define PS2_CLK      9
#endif


enum { 	RAS_I=0, RKS_I,	RHS_I, LAS_I, LKS_I, LHS_I, TRTS_I };        // Define the order of our servos...
//Servo     g_aServos[MNUMSERVOS];    // Define servos object for each servo

//--------------------------------------------------------------------------
// Try to define our own Servo Group move code class
//--------------------------------------------------------------------------


// Also define the scheduler to handle the group servo move...
// The scheduler.
//Scheduler g_sched = Scheduler();        // Used to schedule background tasks
//Task*    g_ptaskSGM;                     // task for handling the group move - synchronize the move of servos



typedef struct _brat_save_data
{
    byte	cServos;		// count of servos in our save data.
    byte	bChksum;		// Checksum.  simple check to see if the data is valid
    struct _asd				// Servo Data
    {
        short	sPWOffset;	        // The PW Offset in microseconds.
    } asd[MNUMSERVOS];

} BRAT_SAVE_DATA;



long         g_lWalkSpeed1;
long         g_lWalkAngle1;
short        g_sLastStep;
short        g_sAnkleAdj;
short        g_sTravLength;
boolean      g_fBotActive;
boolean      g_fServosInit;

const CMLSD1 *g_pcmlsd1;				// Current command list that is active
short	g_imlsd1;				// Current index to which command in the command list is active.
#define SET_MLSD1(pm,ra,rk,rh,la,lk,lh, time)  \
	(pm)->rightanklepos = (ra);	\
	(pm)->rightkneepos = (rk);	\
	(pm)->righthippos = (rh);		\
	(pm)->leftanklepos = (la);	\
	(pm)->leftkneepos = (lk); 	\
	(pm)->lefthippos = (lh);		\
	(pm)->sMoveTime = (time);
	

boolean g_fShowDebugPrompt;
boolean g_fDebugOutput;
#define TravelDeadZone 5

//--------------------------------------------------------------------------
// define some forward references
//--------------------------------------------------------------------------
extern short GetIRScanInCM(byte channel);
//extern void FindServoZeroPoints(SERVO_LIST *pservos);
//extern void MoveLegServosDegrees1 (short rightanklepos,short rightkneepos,short righthippos,short leftanklepos,short leftkneepos,short lefthippos,short sMoveTime);
extern boolean FProcessCurrentCommandList(void);
extern boolean ClockHasElapsed(unsigned long ulTimeRef, unsigned long ulDT);
#ifdef PS2_TEST
extern void PS2_Test(void);
#endif
#define SERVO_TEST
#ifdef SERVO_TEST
extern void ServoRangeTest(void);
#endif

//--------------------------------------------------------------------------
// define the movements for our standard commands...
//--------------------------------------------------------------------------
MLSD1 g_mlsd1_Walk[] = {{0,0,0,0,0,0,500}, {0,0,0,0,0,0,500}};
static const CMLSD1 g_cmlsd1_Walk PROGMEM = {2, 0};    // we are not going to use the pointer here...



//--------------------------------------------------------------------------
// SETUP: the main arduino setup function.
//--------------------------------------------------------------------------
void setup(){
    int error;
    
    g_pcmlsd1 = 0;		// Current command list that is active
    g_imlsd1 = 0;		// Current index to which command in the command list is active.
    g_tcPausePeriod = 0; // make sure we don't start off with some delay defined
    g_tcLastActive = micros();	// remember our time
    g_fDebugOutput = false;			// start with it off!
    g_fShowDebugPrompt = true;
    g_sTravLength = 5;				// default for ps2...
    g_fServosInit = false;
    
   delay(250);
    Serial.begin(57600);

    Serial.println("Botboardduino Brat Program Startup");

    //error = ps2x.config_gamepad(57, 55, 56, 54);  // Setup gamepad (clock, command, attention, data) pins
    error = ps2x.config_gamepad(PS2_CLK, PS2_CMD, PS2_SEL, PS2_DAT);  // Setup gamepad (clock, command, attention, data) pins
 
    if(error == 0){
       Serial.println("Found Controller, configured successful");
    }
   
     else if(error == 1)
       Serial.println("No controller found");
   
    else if(error == 2)
       Serial.println("Controller found but not accepting commands.");
 
}


//--------------------------------------------------------------------------
// Loop: the main arduino main Loop function
//--------------------------------------------------------------------------

void loop(){
    // We also have a simple debug monitor that allows us to 
    // check things. call it here..
    if (TerminalMonitor())
        return;           
        
    CheckVoltages();    // check voltages - if we get too low shut down the servos...        

    // call off to process our current command list (if any).  It will return TRUE if a command is still active
    // this will influence which commands we may process while a command is still active.
    g_fCommandActive = FProcessCurrentCommandList();	
    unsigned long tcRet;

    if (mservo.WhichGroupMoveServosActive())		// check to see if any group moves are going on...
        g_tcLastActive = micros();			// remember our time

    // For now start off only with PS2 data
    tcRet = CheckAndProcessPS2Data();

    delay(tcRet);
      
}

//==============================================================================
// InitializeServos - We will delay when we call the servo init stuff until a
//        command such as start is called on.  This should make it that if your
//        robot starts off running on USB power it does not overpower it...
//==============================================================================
void InitializeServos(void)
{
    if (g_fServosInit)
        return;
    g_fServosInit = true;
    
    // Init all of the servos - Should setup a better way...
    mservo.AttachServo(RAS_I, RAS_PIN, false);
    mservo.AttachServo(RKS_I, RKS_PIN, false);
    mservo.AttachServo(RHS_I, RHS_PIN, false);
    mservo.AttachServo(LAS_I, LAS_PIN, true);
    mservo.AttachServo(LKS_I, LKS_PIN, true);
    mservo.AttachServo(LHS_I, LHS_PIN, true);
    mservo.AttachServo(TRTS_I, TRTS_PIN, false);
    LoadServosConfig();       // reload any offsets that we may have stored away
}

//==============================================================================
// CheckVoltages: Try to catch when the voltages drop too low and turn off the
// servos.  This can be because the battery level got too low or if we turn it off
//     and the Arduino is still being powered by USB
//==============================================================================
word g_wVSPrev;

void CheckVoltages(void) 
{
#ifdef AVS_PIN
    word wVS = analogRead(AVS_PIN);
#ifdef DEBUG
    if (g_fDebugOutput && (wVS != g_wVSPrev)) {
        Serial.print("VS: ");
        Serial.println(wVS, DEC);
        g_wVSPrev = wVS;
    }
#endif
    if ((wVS < (word)AVAL_MIN) && g_fServosInit) {
        mservo.DetachServo(RAS_I);
        mservo.DetachServo(RKS_I);
        mservo.DetachServo(RHS_I);
        mservo.DetachServo(LAS_I);
        mservo.DetachServo(LKS_I);
        mservo.DetachServo(LHS_I);
        mservo.DetachServo(TRTS_I);
        g_fServosInit = false;
        g_fBotActive = false;
        MSound(SOUND_PIN, 3, 100, 2500, 100, 2500, 100, 2500);
    }
#endif
}
    
    
    
    


//==============================================================================
// CheckAndProcessPS2Data -This is code the checks for and processes input from
//        the PS2 controller.
//==============================================================================
unsigned long CheckAndProcessPS2Data(void)
{
    boolean fPacketChanged;
    short	JoyStickValueIn;
	
#ifdef DEBUG_ENTERLEAVE
    Serial.print("Enter: CheckAndProcessPS2DATA: ");
    Serial.print(micros(), DEC);
    Serial.println("");
#endif
    // Then try to receive a packet of information from the PS2.
    ps2x.read_gamepad();          //read controller and set large motor to spin at 'vibrate' speed

    // Wish the library had a valid way to verify that the read_gamepad succeeded... Will hack for now
    if ((ps2x.Analog(1) & 0xf0) != 0x70) {
        // Controller did not return Analog mode...
        Serial << "PS2 - error (" << _HEX( ps2x.Analog(1)) << "): " << _HEX( ps2x.Analog(3)) << " "  << _HEX( ps2x.Analog(4)) << " " 
                << _HEX( ps2x.Analog(5)) << " " << _HEX( ps2x.Analog(6)) << " " << _HEX( ps2x.Analog(7)) << " " << _HEX( ps2x.Analog(8)) << endl;
        return 100;
    }

    if(ps2x.ButtonPressed(PSB_START)) {		// OK lets try "0" button for Start. 
        if(g_fBotActive) {
	    //Turn off
            MSound(SOUND_PIN, 3, 100, 2500, 80, 2250, 60, 2000);
	    mservo.DumpTimings();

	    SetCurrentCommandList(&s_cmlsd1_Rest);		//command = 9;
	    g_fBotActive = false;
	} else 	{
	    //Turn on
            MSound(SOUND_PIN, 3, 60, 2000, 80, 2250, 100, 2500);
            InitializeServos();                                // Make sure the servos are active
	    SetCurrentCommandList(&s_cmlsd1_Home);		//command = 0
	    g_fBotActive = true;
	}
    }	

    if (g_fBotActive) {
	if (ps2x.ButtonPressed(PSB_PAD_UP))
	    SetCurrentCommandList(&s_cmlsd1_UpFromFront);		//command = 7

	else if (ps2x.ButtonPressed(PSB_PAD_DOWN))
	    SetCurrentCommandList(&s_cmlsd1_UpFromBack);		//command = 8
		
	else if (ps2x.ButtonPressed(PSB_BLUE))				// X button test
	    SetCurrentCommandList(&s_cmlsd1_Home);				//command = 0
	
	else if (ps2x.ButtonPressed(PSB_SELECT)) {		// Select button test
	    // Can adjust the travel length...
	    if (g_sTravLength == 5)
	        g_sTravLength = 8;
	    else
		g_sTravLength = 5;
        }
	
        // Commands that are set by joystick should only be set when a command is not currently active.	
	else if (!g_fCommandActive) {
	    if (abs(ps2x.Analog(PSS_RY)-128) > TravelDeadZone)  {
				 
	        // Note, Speed and Angle here are in 10s so multiply by 10...		
		JoyStickValueIn = ps2x.Analog(PSS_RY) - 128;
		if (JoyStickValueIn > 0) {
		    g_lWalkSpeed1 = (JoyStickValueIn*10)/g_sTravLength;
		    g_lWalkAngle1 = (JoyStickValueIn + 40) - g_sAnkleAdj;
		} else 	{
		    g_lWalkSpeed1 = (JoyStickValueIn*10)/g_sTravLength;
		    g_lWalkAngle1 = -((JoyStickValueIn - 40) + g_sAnkleAdj);
		}

	        // We are setting up to do our main walk forward or backward.

		if (g_sLastStep == 0) {
			SET_MLSD1(&g_mlsd1_Walk[0],g_lWalkAngle1, g_lWalkSpeed1, g_lWalkSpeed1, -g_lWalkAngle1,-g_lWalkSpeed1,-g_lWalkSpeed1,500);
			SET_MLSD1(&g_mlsd1_Walk[1],-g_lWalkAngle1, g_lWalkSpeed1, g_lWalkSpeed1,  g_lWalkAngle1,-g_lWalkSpeed1,-g_lWalkSpeed1,500);
			g_sLastStep = 1;
	        } else if (g_sLastStep == 1) {
			SET_MLSD1(&g_mlsd1_Walk[0], -g_lWalkAngle1,-g_lWalkSpeed1,-g_lWalkSpeed1,  g_lWalkAngle1, g_lWalkSpeed1, g_lWalkSpeed1,500); 
			SET_MLSD1(&g_mlsd1_Walk[1], g_lWalkAngle1,-g_lWalkSpeed1,-g_lWalkSpeed1, -g_lWalkAngle1, g_lWalkSpeed1, g_lWalkSpeed1,500); 
			g_sLastStep = 0;
		}
		g_sAnkleAdj = g_sAnkleAdj + 10;
		if (g_sAnkleAdj > 60)
			g_sAnkleAdj = 60;
		
		SetCurrentCommandList(&g_cmlsd1_Walk);		//command = 1

	    } else if (abs(ps2x.Analog(PSS_RX)-128) > TravelDeadZone) {
		JoyStickValueIn = ps2x.Analog(PSS_RX) - 128;
		if ((JoyStickValueIn > TravelDeadZone) && (JoyStickValueIn < 70))
		    SetCurrentCommandList(&s_cmlsd1_CCRight);		//command = 14
		else if (JoyStickValueIn >= 70)
		    SetCurrentCommandList(&s_cmlsd1_TurnRight);		//command = 12
		else if ((JoyStickValueIn < -TravelDeadZone) && (JoyStickValueIn > -70))
		    SetCurrentCommandList(&s_cmlsd1_CCLeft);		//command = 13
		else
		    SetCurrentCommandList(&s_cmlsd1_TurnLeft);		//command = 11
	    }
			
	    // BUGBUG: We will allow the other right UP/Down joystick change the hip angles to allow us to 
	    // shoot.  Currenly only when we are not walking, but may change to instead of setting an independent
	    // value we will instead build an offset from the movements.  But would probably need a button to press
	    // to go back to zero...
            if ((mservo.WhichGroupMoveServosActive() & ((1 << RHS_I) | (1 << LHS_I))) == 0) {
    
                if (abs(ps2x.Analog(PSS_LY)-128) > TravelDeadZone) {
                    g_RHipServoMS += (short)(ps2x.Analog(PSS_LY)-128);
    		    if (g_RHipServoMS > 2000)
    		        g_RHipServoMS = 2000;
        	    else if (g_RHipServoMS < 1000 )
    		        g_RHipServoMS = 1000;
                    g_LHipServoMS += (short)(ps2x.Analog(PSS_LY)-128);
                    if (g_LHipServoMS > 2000)
                        g_LHipServoMS = 2000;
                    else if (g_LHipServoMS < 1000 )
                        g_LHipServoMS = 1000;
                    mservo.GroupMoveStart();
                    mservo.ServoMoveMicroseconds(RHS_I, g_RHipServoMS);
                    mservo.ServoMoveMicroseconds(LHS_I, g_LHipServoMS );
                    mservo.GroupMoveCommit(250);
                }
	    }
        }
	//
	// some commands can be processed in the secondary loop...
	//
        if ((mservo.WhichGroupMoveServosActive() & (1 << TRTS_I)) == 0) {

            if (abs(ps2x.Analog(PSS_LX)-128) > TravelDeadZone) {
                word wTurretMS = mservo.ServoReadMicroseconds(TRTS_I) + (short)(ps2x.Analog(PSS_LX)-128);
		if (wTurretMS > 2000)
                    wTurretMS = 2000;
                else if (wTurretMS < 1000 )
                    wTurretMS = 1000;
                if (wTurretMS != mservo.ServoReadMicroseconds(TRTS_I)) {
                    mservo.GroupMoveStart();
                    mservo.ServoMoveMicroseconds(TRTS_I, wTurretMS);
                    mservo.GroupMoveCommit(250);
                }
            }
        }

	// Check to see if we are asked for distance - test of bidirectional sending of data
	if (ps2x.ButtonPressed(PSB_RED)) {			// O button test
	    // OK lets do a distance 
#ifdef LATER
	    byte b = GetIRScanInCM(IRPIN);		// get it into some dummy variable.
				
	    // then tell remote to display the value for us.
	    // We will use simple version where we pass the number(byte) to remote as the equiv of packet number
	    // but later will add capabilities to pass strings as extra bytes...
#ifdef DEBUG

            ich = sprintf((char*)szTemp, "Distance in cm: %d\n", b);
            Serial.write(szTemp, ich);

#endif
	    XBeeOutputVal(b);
#endif // LATER
        }

#ifdef DEBUG
	if (ps2x.ButtonPressed(PSB_PINK)) {	// Square button
	    if (g_fDebugOutput)
		g_fDebugOutput = false;		// turn off debug outputs
	    else
		g_fDebugOutput = true;		// Turn it back on.
	}
#endif
	
    }

    // If we are not active and a timer has elapsed then go into low power mode...
    // go into the low power mode...
    if (g_fBotActive && !g_fCommandActive) {			// only do this if we are actually still active and no command active...
	if (ClockHasElapsed(g_tcLastActive, (unsigned long)10000000)) {	//10 seconds
	    SetCurrentCommandList(&s_cmlsd1_Rest);		//command = 9
	    g_tcLastActive = micros();			// remember our time - so won't repeat for another 10 seconds...
	} else if (ClockHasElapsed(g_tcLastActive, (unsigned long) 500000))	//.5 seconds
	    g_sAnkleAdj = 0;		// this will happen everytime through from .5 to 10 seconds, but will cost as much to avoid...
    }

    // Return to ok to give us a little break;
#ifdef DEBUG_ENTERLEAVE
    Serial.print("Leave: CheckAndProcessPS2DATA: ");
    Serial.print(micros(), DEC);
    Serial.println("");
#endif

    return 20;
}

//==============================================================================
// SetCurrentCommandList; simple helper function that sets the current command list
//==============================================================================
boolean SetCurrentCommandList(const CMLSD1 *pcmlsd1)
{
#ifdef DEBUG
    // setup to output back to our USB port
    if (g_fDebugOutput) {
        unsigned long tc =  micros();
	Serial << "\nSetCurrentCommandList(" << _DEC( tc - g_tcLastDebug) << ") : " << _HEX((unsigned short)pcmlsd1 & 0xffff) << " " <<
                _HEX(mservo.WhichGroupMoveServosActive()) << endl;
        g_tcLastDebug = tc;        
    }
#endif
    g_pcmlsd1 = pcmlsd1;
    g_imlsd1 = -1;					// setup count of the item list to -1 F Process will handle it...
    g_fCommandActive = true;		// make sure we don't time out...
    return true;
}


//==============================================================================
// FProcessCurrentCommandList - Main function that does a query to see if the 
// 		multiple servo code needs to update.  If a command finishes it then sees if
//		there are more commands to process if so starts up the next one...
//==============================================================================
boolean FProcessCurrentCommandList(void)
{
    if (!g_pcmlsd1)
        return false;		// no current command

    // Then see if our leg servos are still active
    if (mservo.WhichGroupMoveServosActive() & 0x3f)
        return true;		// at least one of our leg servos is still moving...

    // Or see if we have some delay defined
    if (g_tcPausePeriod) {
        if (!ClockHasElapsed(g_tcPauseStart, g_tcPausePeriod))
            return true;		// nope - say a command is still active
        g_tcPausePeriod = 0;		// clear it out.
    }

    g_imlsd1++;				// inrement to the next command in our list (note started at minus one so should be fine
    if (g_imlsd1 < pgm_read_byte(&g_pcmlsd1->cItems)) {
        // we need to convert each of these values to drive speed
        MLSD1 mlsd1;
        
        // OK we need to grab the next command from either our space or program space...
        if (g_pcmlsd1 == &g_cmlsd1_Walk)   // BUGBUG:: should be cleaner way.  Could also special case everywhere below..
            mlsd1 = g_mlsd1_Walk[g_imlsd1]; //memcpy(&mlsd1, &(g_mlsd1_Walk[g_imlsd1]), sizeof(MLSD1));
        else {
            MLSD1 *pmlsd1 = ((MLSD1*)pgm_read_word(&(g_pcmlsd1->pamlsd1))) + g_imlsd1;
            memcpy_P(&mlsd1, pmlsd1, sizeof(MLSD1));
        }    
#ifdef DEBUG
        if (g_fDebugOutput) {
        // setup to output back to our USB port
            unsigned long tc =  micros();
            Serial << "M(" << _DEC(tc - g_tcLastDebug) << ") " << _HEX(mservo.WhichGroupMoveServosActive()) 
                    << ": RAKH: " << _DEC(mlsd1.rightanklepos) << " " << _DEC(mlsd1.rightkneepos) << " " << _DEC(mlsd1.righthippos) 
                    << ": L: " << _DEC(mlsd1.leftanklepos) << " " << _DEC(mlsd1.leftkneepos) << " " << _DEC(mlsd1.lefthippos)
                    << " t:" << _DEC(mlsd1.sMoveTime) << endl; 
            g_tcLastDebug = tc;
    	}
#endif
        // Setup a special case.  If the move time is less than zero than this is a pause for the positive amount
        // of time, not an actual group move item.
        if (mlsd1.sMoveTime < 0 ) {
            g_tcPauseStart =  micros();					// remember when we started this delay
            g_tcPausePeriod = ((long)-(mlsd1.sMoveTime)) * 1000;	// convert to us
        } else 	{
            g_tcPausePeriod = 0;	// clear the pause period so we won't use it anymore...
            mservo.GroupMoveStart();
            mservo.ServoMoveDegrees1(RAS_I, mlsd1.rightanklepos);
            mservo.ServoMoveDegrees1(RKS_I, mlsd1.rightkneepos);
            mservo.ServoMoveDegrees1(RHS_I, mlsd1.righthippos);
            mservo.ServoMoveDegrees1(LAS_I, mlsd1.leftanklepos);
            mservo.ServoMoveDegrees1(LKS_I, mlsd1.leftkneepos);
            mservo.ServoMoveDegrees1(LHS_I, mlsd1.lefthippos);
            mservo.GroupMoveCommit(mlsd1.sMoveTime);
        }
        return true;
    }

    g_pcmlsd1 = 0;		// Null out the list
    g_LHipServoMS = mservo.ServoReadMicroseconds(LHS_I);
    g_RHipServoMS = mservo.ServoReadMicroseconds(RHS_I);
    return false;		// we processed our whole command list
}		


//==============================================================================
// TerminalMonitor - Simple background task checks to see if the user is asking
//    us to do anything, like update debug levels ore the like.
//==============================================================================
boolean TerminalMonitor(void)
{
    byte szCmdLine[20];
    int ich;
    int ch;
    // See if we need to output a prompt.
    if (g_fShowDebugPrompt) {
        Serial.println("Arduino Brat Monitor");
        Serial.println("D - Toggle debug on or off");
        Serial.println("O - Enter Servo offset mode");
        Serial.println("S <SN> <ANGLE> - Move Servo");
        g_fShowDebugPrompt = false;
    }
       
    // First check to see if there is any characters to process.
    if (ich = Serial.available()) {
        ich = 0;
        // For now assume we receive a packet of data from serial monitor, as the user has
        // to click the send button...
        for (ich=0; ich < sizeof(szCmdLine); ich++) {
            ch = Serial.read();        // get the next character
            if ((ch == -1) || ((ch >= 10) && (ch <= 15)))
                break;
             szCmdLine[ich] = ch;
        }
        szCmdLine[ich] = '\0';    // go ahead and null terminate it...
        Serial.print("Serial Cmd Line:");        
        Serial.write(szCmdLine, ich);
        Serial.println("!!!");
        
        // So see what are command is.
        if (ich == 0) {
            g_fShowDebugPrompt = true;
        } else if ((ich == 1) && ((szCmdLine[0] == 'd') || (szCmdLine[0] == 'D'))) {
            g_fDebugOutput = !g_fDebugOutput;
            if (g_fDebugOutput) 
                Serial.println("Debug is on");
            else
                Serial.println("Debug is off");
        } else if ((ich == 1) && ((szCmdLine[0] == 'o') || (szCmdLine[0] == 'O'))) {
            FindServoZeroPoints();
        } else if (((szCmdLine[0] == 's') || (szCmdLine[0] == 'S'))) {
            // ok lets grab a servo number
            int iServo = 0;
            short sSign = 1;
            int iAngle = 0;    
            int i;        
            for (i=1; i< ich; i++) {
                if (szCmdLine[i] != ' ')
                    break;
            }
            
            if ((szCmdLine[i] >= '0') && (szCmdLine[i] <= '9')) {
                iServo = szCmdLine[i++] - '0';
                
                // now angle
                while (szCmdLine[i] == ' ')
                    i++;
                if (szCmdLine[i] == '-') {
                    sSign = -1;
                    i++;
                }
                while ((szCmdLine[i] >= '0') && (szCmdLine[i] <= '9')) {
                    iAngle = iAngle*10 + szCmdLine[i++] - '0';
                }
                // Ok lets try moving the servo there...
                InitializeServos();        // Make sure the servos have been initialized...
                mservo.MoveOneServoDegrees1(iServo, sSign * iAngle, 500);
            }
        }
        
        return true;
    }
    return false;
} 



//==============================================================================
// ClockHasElapsed: Pass in a time value and a dt and it checks to see
//         if that amount of time has elapsed.
//==============================================================================
boolean ClockHasElapsed(unsigned long ulTimeRef, unsigned long ulDT)
{
    if ((micros() - ulTimeRef) > ulDT)
        return true;
    else
        return false;
}

//==============================================================================
//[GetIRSensorInCM] - Get the distance to a target using a Sharp IR sensor.  These
// devices are not linear so we use a look up table to convert the results to CM.
// Parameters: IRPIN - Which pin to use
// Return: Distance in CM.  Range is from 0 to 80
//==============================================================================
// Define the table of values to convert
// This table is used to convert Sharp IR readings into CM
static const byte  s_abIRCMTable[] = {80,80,80,80,80,80,80,80,80,78,
                        76,74,72,70,68,66,64,62,60,59,
                        58,57,55,53,52,51,50,49,48,47,
                        45,43,42,41,40,39,38,37,35,33, 
                        32,31,30,30,29,29,28,28,27,27, 
                        26,26,26,25,25,25,24,24,24,23, 
                        23,22,22,21,21,20,20,20,19,19, 
                        18,18,18,17,17,16,16,16,15,15, 
                        15,14,14,13,13,13,12,12,11,11, 
                        11,10,10,10,10,10,10,10,10,10 };



short GetIRScanInCM(byte  channel)
{                        
  	// first read in the IRvalue.
	word wScanRange = analogRead(channel);

  	// If it is greater than 512 it is out of range we will return 0
	if (wScanRange > 512)		// outside our table so return 0 to signal too close 
		return 0;

	// now convert entry to cm from table.  Old program
	// divided value by 5.12 using floating point, but could simply multiply by 100 and divide by 512
	// in integer math, which should be a lot quicker.  512*100 < 65536 so fits in word...
	return (short) &s_abIRCMTable[(wScanRange*100)/512];
}



//==============================================================================
// MoveLegServosDegrees1: simple helper function to convert the six servos from degrees to pulse width units and then call off to my
// 				groupmove function.
//==============================================================================
void MoveLegServosDegrees1 (short rightanklepos,short rightkneepos,short righthippos,short leftanklepos,short leftkneepos,short lefthippos,short sMoveTime)
{
    // we need to convert each of these values to drive speed
    word awTo[6];

#ifdef DEBUG
    if (g_fDebugOutput) {
	// setup to output back to our USB port
	unsigned long tc =  micros();
        Serial << "M:(" << _DEC(tc - g_tcLastDebug) << ") " << _HEX(mservo.WhichGroupMoveServosActive()) 
                << ": RAKH: " << _DEC(rightanklepos) << " " << _DEC(rightkneepos) << " " << _DEC(righthippos) 
                << ": L: " << _DEC(leftanklepos) << " " << _DEC(leftkneepos) << " " << _DEC(lefthippos)
                << " t:" << _DEC(sMoveTime) << endl; 

	g_tcLastDebug = tc;
    }
#endif

    mservo.GroupMoveStart();
    mservo.ServoMoveDegrees1(RAS_I, rightanklepos);
    mservo.ServoMoveDegrees1(RKS_I, rightkneepos);
    mservo.ServoMoveDegrees1(RHS_I, righthippos);
    mservo.ServoMoveDegrees1(LAS_I, leftanklepos);
    mservo.ServoMoveDegrees1(LKS_I, leftkneepos);
    mservo.ServoMoveDegrees1(LHS_I, lefthippos);
    mservo.GroupMoveCommit(sMoveTime);
}



//==============================================================================
//	LoadServosConfig - We will try to restore the servo config information from
//		the eeprom on the board.  We will store some checksum information to
//		help us verify that the information is correct.
// 		Will use the new 
//==============================================================================
void LoadServosConfig()
{
    BRAT_SAVE_DATA bsd;
    byte bChksum = 0;
    byte i;
    byte *pb;
    boolean fValid = true;

    // Now try to read the block out starting at 0...
    pb = (byte*)&bsd;
    for (i=0; i< sizeof(bsd); i++) {
       *pb++ = EEPROM.read(i);
    }

    // Make sure the count of servos looks about right...
    if ((bsd.cServos >=6) && (bsd.cServos <= MNUMSERVOS)) {
        // lets first loop through and calculate the checksum to make
        // sure the data looks valid.
        for (i=0; i< bsd.cServos; i++) 	{
            bChksum += (bsd.asd[i].sPWOffset & 0xff);
            bChksum += bsd.asd[i].sPWOffset >> 8;
            // do a little pre-validation...    

            if ((bsd.asd[i].sPWOffset < -200) || (bsd.asd[i].sPWOffset > 200))
                fValid = false;
        }	

        if (fValid && (bChksum == bsd.bChksum)) {
            // we have valid data, update our servos...
            for (i=0; i< bsd.cServos; i++) {
                mservo.SetServoOffset(i, bsd.asd[i].sPWOffset);
            }
        }
    }

    // Lets dump the servo offsets to see what we have...
#ifdef DEBUG
    if (g_fDebugOutput) {
        Serial.print("Load Servo offsets :");
        Serial.print(fValid, HEX);
        // we have valid data, update our servos...
        for (i=0; i<MNUMSERVOS; i++) {
            Serial.print(" ");
            Serial.print(mservo.GetServoOffset(i), DEC);
        }
        Serial.println("");
    }
#endif        
}




//==============================================================================
//	FindServoZeroPoints - Find the zero points for each of our servos... 
// 		Will use the new servo function to set the actual pwm rate and see
//		how well that works...
//==============================================================================
void FindServoZeroPoints()
{
    // not clean but...
    static char *apszServos[] = {"RA","RK","RH", "LA", "LK", "LH", "HD"};
    int data;
    short sSN = 0; 			// which servo number
    boolean fNew = true;	// is this a new servo to work with?
    boolean fExit = false;	// when to exit
    int ich;
    short sOffset;
    
    // OK lets move all of the servos to their zero point.
    InitializeServos();
    Serial.println("Find Servo Zeros.\n$-Exit, +- changes, *-change servo");
    MoveLegServosDegrees1 (0, 0, 0, 0, 0, 0, 0);


    while(!fExit) {
        if (fNew) {
            sOffset = mservo.GetServoOffset(sSN);
            Serial << "Servo: " <<  apszServos[sSN] << "(" << _DEC(sOffset) << ")" << endl; 

	    // Now lets wiggle the servo
            mservo.WaitForGMToComplete(0x3f);
            mservo.MoveOneServoDegrees1(sSN, 250, 500);

            mservo.WaitForGMToComplete(0x3f);
            mservo.MoveOneServoDegrees1(sSN, -250, 500);

            mservo.WaitForGMToComplete(0x3f);
            mservo.MoveOneServoDegrees1(sSN, 0, 500);
            fNew = false;
        }
	
	//get user entered data
	data = Serial.read();
	//if data received
	if (data !=-1) 	{
            if (data == '$')
		fExit = true;	// not sure how the keypad will map so give NL, CR, LF... all implies exit

	    else if ((data == '+') || (data == '-')) {
	        if (data == '+')
		    sOffset += 5;		// increment by 5us
		else
		    sOffset -= 5;		// increment by 5us

		Serial <<"    " << _DEC(sOffset) << endl; 
                mservo.SetServoOffset(sSN, sOffset);
                mservo.MoveOneServoDegrees1(sSN, 0, 50);
	    } else if ((data >= '0') && (data <= '6')) {
		// direct enter of which servo to change
		fNew = true;
		sSN = data - '0';
	    } else if (data == '*') {
	        // direct enter of which servo to change
		fNew = true;
		sSN++;
		if (sSN == MNUMSERVOS) 
		    sSN = 0;	
	    }
	}
    }
    Serial.print("Find Servo exit ");
    for (sSN=0; sSN < MNUMSERVOS; sSN++){
        Serial.print(" ");
        Serial.print(mservo.GetServo(sSN)->readMicroseconds(), DEC);
    }

    Serial.print("\nSave Changes? Y/N: ");

    //get user entered data
    while (((data = Serial.read()) == -1) || ((data >= 10) && (data <= 15)))
	; 

    if ((data == 'Y') || (data == 'y')) {
        // Ok they asked for the data to be saved.  We will store the data with a 
        // number of servos (byte)at the start, followed by a byte for a checksum...followed by word for each servo center and for the
	// heck of it a word for the range, although we don't muck with it yet, we might...
	// 
	BRAT_SAVE_DATA bsd;
	bsd.cServos = MNUMSERVOS;
	bsd.bChksum = 0;
	for (sSN=0; sSN < MNUMSERVOS; sSN++) {
	    bsd.asd[sSN].sPWOffset =  mservo.GetServoOffset(sSN);
	    bsd.bChksum += (bsd.asd[sSN].sPWOffset & 0xff);
	    bsd.bChksum += bsd.asd[sSN].sPWOffset >> 8;
	}

	// Will write the block out starting at 0... may later change to use some of the other stuff.
        byte *pb = (byte*)&bsd;
        
        for (sSN=0; sSN < sizeof(bsd); sSN++) 
            EEPROM.write(sSN, *pb++);
    } else {
        void LoadServosConfig();
    }

}


//==============================================================================
//	SoundNoTimer - Quick and dirty tone function to try to output a frequency
//            to a speaker for some simple sounds.
//==============================================================================
void SoundNoTimer(uint8_t _pin, unsigned long duration,  unsigned int frequency)
{
    volatile uint8_t *pin_port;
    volatile uint8_t pin_mask;
    long toggle_count = 0;
    long lusDelayPerHalfCycle;
    
    // Set the pinMode as OUTPUT
    pinMode(_pin, OUTPUT);

    pin_port = portOutputRegister(digitalPinToPort(_pin));
    pin_mask = digitalPinToBitMask(_pin);
    
    toggle_count = 2 * frequency * duration / 1000;
    lusDelayPerHalfCycle = 1000000L/(frequency * 2);
    
    // if we are using an 8 bit timer, scan through prescalars to find the best fit
    while (toggle_count--) {
        // toggle the pin
        *pin_port ^= pin_mask;
        
        // delay a half cycle
        delayMicroseconds(lusDelayPerHalfCycle);
    }    
    *pin_port &= ~(pin_mask);  // keep pin low after stop

}

void MSound(uint8_t _pin, byte cNotes, ...)
{
	va_list ap;
	unsigned int uDur;
	unsigned int uFreq;
	va_start(ap, cNotes);

	while (cNotes > 0) {
		uDur = va_arg(ap, unsigned int);
		uFreq = va_arg(ap, unsigned int);
		SoundNoTimer(_pin, uDur, uFreq);
		cNotes--;
	}
	va_end(ap);
}


