//=============================================================================
// Movement.c - Support file to do some simple servo movments for DeBrat 
//==============================================================================
//#define DEBUG
//#define DEBUG_VERBOSE
//#define DEBUG_TIMING



#include <avr/interrupt.h>
#if ARDUINO>99
#include <Arduino.h> // Arduino 1.0
#else
#include <Wprogram.h> // Arduino 0022
#endif
#include "ServoGroupMove.h"
extern boolean g_fDebugOutput;


// Number of ms between timer overlfows
//#define DBG_PIN  59
#define MSPERSGMCALLBACK  16					// BUGBUG:: Set to ms will change to US ..hard coded may try to set depending of differences

//static MSERVODATA s_msd = {0, 0};

#ifdef DEBUG_TIMING
#define 	TCDEBUG_SIZE 200
unsigned	g_ctcDebug;
unsigned long	g_atcDebug[TCDEBUG_SIZE];

#endif

// Define our one instance of the class:
MServo mservo;
//==============================================================================
// This is the main Callback function that is called to update the servo positions
// for the group moves.  This function is a callback function that is called by
// the background scheduler.
//==============================================================================
//--------------------------------------------------------------
// Timer 2 interrupt
//  Called when Timer 2 interrupts. Finds tasks to execute,
//  and execute them.
//--------------------------------------------------------------
#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega328P__) || defined(__AVR_ATmega1280__)
ISR(TIMER2_COMPA_vect)
#else
ISR(TIMER2_COMP_vect)
#endif
{
    PMSERVOITEM psrvi;
    short i;
    unsigned uMask;				// mask for which servo number
    unsigned uStillActive;		// Bits for which servos have work still to go
    short sDSPrev;

    // This is an interrupt handler, so interrupts will be disabled when we come in.
    // Lets reanable quickly to allow other things to continue
    TCNT2 = 0;
  
    sei();    // enable interrupts now.

    // first see if we have anything to do.
    if (mservo._uWhichServosActive) {
#ifdef DEBUG_INT_HANDLER
        if (g_fDebugOutput) {
            ich = sprintf((char*)szTemp, "FPGM: %x\n", mservo._uWhichServosActive);
            Serial.write(szTemp, ich);
        }
#endif
#ifdef DEBUG_TIMING
	if (g_ctcDebug < TCDEBUG_SIZE)
		g_atcDebug[g_ctcDebug++] = micros();
#endif
	uStillActive = 0;				// 
	uMask = 0x1;					// set mask
	for (i=0;i< NUM_GROUP_MOVE_SERVOS;i++) {
	    // now increment each servo
	    psrvi = &mservo._asrvi[i];						// pointer to the data associated with this servo

	    // Since this  happens on an interrupt basis, only update thos items that are in the mask of items
	    // this will keep us from having our interrupt start processing a group move, until the complete
	    // update has been completed.
	    if ((mservo._uWhichServosActive & uMask) && psrvi->cCyclesLeft) {
		// Ok this one has something to do
		if (--(psrvi->cCyclesLeft)) {		
			psrvi->servo.writeMicroseconds(psrvi->servo.readMicroseconds() + psrvi->sDeltaPW);
		    uStillActive |= uMask;					// Set the bit for this IO pin to say it is still active...
		} else
            psrvi->servo.writeMicroseconds(psrvi->wPWTo);
	    }
	    uMask <<= 1;					// go up to the next bit.
	}	
	
	mservo._uWhichServosActive = uStillActive;		// remember if there is still stuff left to do
    }

    // If there is still group moves going on, we will rescheule ourself.
    mservo._cSGMCallback++;
    if (mservo._uWhichServosActive == 0) {
#ifdef DEBUG_INT_HANDLER
        if (g_fDebugOutput) {
            ich = sprintf((char*)szTemp, "FPGM-Done\n");
            Serial.write(szTemp, ich);
        }
#endif
        mservo.DisableTimer();
//        g_ptaskSGM->repeats(false);
    }
//    digitalWrite(DBG_PIN, LOW);
}

//==============================================================================
//[MServo::MServo] - Class constructor for MServo class
//==============================================================================
MServo::MServo()
{
    _cSGMCallback = 0;
#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega328P__) || defined(__AVR_ATmega1280__)
    TCCR2A = (1 << WGM21);              // CTC
    TCCR2B = (1 << CS22) | (1 << CS21) | (1 << CS20); // set to clock /1024
    OCR2A = 249;                        // overflow every 16ms
#else
    // Use normal mode
    TCCR2 = (1 << WGM21);               // CTC
    TCCR2 |= (1 << CS22) | (1 << CS21) | (1 << CS20); // set prescaler to 125kHz
    OCR2 = 249;                         // so overflow every 16ms
#endif
    TCNT2 = 0;

 //Make sure all servos are initialize properly.
    int i;
     for (i=0; i < NUM_GROUP_MOVE_SERVOS; i++) {
         _asrvi[i].sPWOffset = 0;
         _asrvi[i].wPWTo = 0;
     }
}


//==============================================================================
//[MServo::AttachServo] - This will assign which pin is associated with one of
//                our logical pin numbers...
//==============================================================================
boolean  MServo::AttachServo(byte bServo, byte iPinNum, boolean fReverse)
{
    if (bServo >= NUM_GROUP_MOVE_SERVOS)
        return false;
    _asrvi[bServo].servo.attach(iPinNum);
    _asrvi[bServo].fReverse = fReverse;        // should we reverse the signals
    return true;
}

//==============================================================================
//[MServo::DetachServo] - Will tell servo system to free this pin up.
//==============================================================================
void  MServo::DetachServo(byte bServo)
{
    if (bServo >= NUM_GROUP_MOVE_SERVOS)
        return;
    _asrvi[bServo].servo.detach();
}


//==============================================================================
//[MServo::GetServo] - Get the Servo object associated with a logical pin number
//==============================================================================
Servo*  MServo::GetServo(byte bServo)
{
    if (bServo >= NUM_GROUP_MOVE_SERVOS)
        return (Servo*)0;
    return &_asrvi[bServo].servo;
}



//==============================================================================
//[MServo::GroupMoveStart] - This will start up a group move 
//==============================================================================
void MServo::GroupMoveStart(void)            // Start a group move
{
    byte i;
    for (i=0; i< NUM_GROUP_MOVE_SERVOS; i++) {
        _awNewPW[i] = 0xffff;        // something we can detect...
    }
}


//==============================================================================
//[MServo::GroupMoveCommit] - This is where the main work happens.  It will
//        setup the timer and the like to make all of the servos move to their
//        desired locations and all complete at the specfied time (more or less)
//==============================================================================
void MServo::GroupMoveCommit(word wMoveTime)    // how long in milliseconds the move should take
{
    short i;
    short cCycles;
    PMSERVOITEM psrvi;
    unsigned uNewTimedServos;
    unsigned uWhichServosActiveBefore;
    short wPWPrev;
#ifdef DEBUG
    byte szTemp[80];
    int ich;
    if (g_fDebugOutput)	{
        byte szTemp[80];
	ich = sprintf((char*)szTemp, "MServo::Commit: %d %d\n", wMoveTime, _cSGMCallback);
        Serial.write(szTemp, ich);
	_cSGMCallback = 0;
    }
#endif
    if (wMoveTime)
	cCycles = (wMoveTime+MSPERSGMCALLBACK-1) / MSPERSGMCALLBACK; 	// try to figure how many cycles this is BUGBUG:: we know all of our times are currently n*50
    else
        cCycles = 0;

    uNewTimedServos = 0;
    for (i = 0; i < NUM_GROUP_MOVE_SERVOS; i++) {
        if (_awNewPW[i] == 0xffff)        // something we can detect...
            continue;                    // don't need to process this one...

       	psrvi = &(mservo._asrvi[i]);		// pointer to the data associated with this servo

	// We will do some of this manipulation within a critical section as to keep the
	// data consistent.  Note: there is an if in this so make sure both paths are handled.
	cli();                                                    // probably can do this a little less hammer...
	wPWPrev =psrvi->servo.readMicroseconds();				// Save away the previous word
	psrvi->wPWTo = _awNewPW[i];						// save away the new DS
	if ((wPWPrev != psrvi->wPWTo) && cCycles) {
	    psrvi->sDeltaPW = (psrvi->wPWTo - wPWPrev) / cCycles;
	    psrvi->cCyclesLeft = cCycles;			
	    sei();
	    uNewTimedServos |= (1 << i);	// set the appropriat bit to true
#ifdef DEBUG
#ifdef DEBUG_VERBOSE
	    if (g_fDebugOutput) {
		ich = sprintf((char*)szTemp, " %d:%d/%d/%d/%d", i, cCycles, wPWPrev, (short)psrvi->wPWTo, (short)psrvi->sDeltaPW);
                Serial.write(szTemp, ich);
            }
#endif	
#endif	
        } else 	{
	    psrvi->cCyclesLeft = 0;
	    sei();
#ifdef DEBUG
#ifdef DEBUG_VERBOSE
    	    if (g_fDebugOutput) {
	        ich = sprintf((char*)szTemp, " %d/IM/%d", i, psrvi->wPWTo);
                Serial.write(szTemp, ich);
            }
#endif
#endif
            // Output to the servo object associated with this item...
            psrvi->servo.writeMicroseconds(psrvi->wPWTo);
        }
    }

    cli();
    // Update the list of which servos have group move data.  Do this under critical section to make sure it keeps consistent. 
    // remember the state of which servos were active before we or it in, such that we know if our main callback function is running or not.
    uWhichServosActiveBefore = mservo._uWhichServosActive;
    mservo._uWhichServosActive |= uNewTimedServos;
    sei();
#ifdef DEBUG
#ifdef DEBUG_VERBOSE
    // setup to output back to our USB port
    if (g_fDebugOutput)
    	Serial.println("");
#endif
#endif
    if ((uWhichServosActiveBefore == 0) &&  mservo._uWhichServosActive) {
        mservo.EnableTimer();
    }

#ifdef DEBUG
    else if (g_fDebugOutput) 	{
	ich = sprintf((char*)szTemp, "SGM: NCB! %x %x %d\n", uWhichServosActiveBefore, mservo._uWhichServosActive, cCycles);
        Serial.write(szTemp, ich);
	_cSGMCallback = 0;
        for (i = 0; i < NUM_GROUP_MOVE_SERVOS; i++) {
            if (_awNewPW[i] == 0xffff)        // something we can detect...
                continue;                    // don't need to process this one...
	    ich = sprintf((char*)szTemp, " %d = %d", i, _awNewPW[i]);
            Serial.write(szTemp, ich);
        }   
	Serial.println("");
    }
#endif
    // Now return the main loop will do the calling out to the processServoGroupMove to monitor and update
    // the actual moves.

}

//==============================================================================
//[MServo::ServoMoveMicroseconds] - Tell a specific servo by index number the
//        new desired pulse width in microseconds.  - This function will also
//        add in any servo offsets that were set for the servo.
//==============================================================================
boolean MServo::ServoMoveMicroseconds(byte bServo, word wPW)    // Desired pulse width in microseconds
{
    if (bServo >= NUM_GROUP_MOVE_SERVOS)
        return false;
    if (_asrvi[bServo].fReverse)
        _awNewPW[bServo] = 3000 - (wPW + _asrvi[bServo].sPWOffset);
    else    
        _awNewPW[bServo] = wPW + _asrvi[bServo].sPWOffset;
    return true;    
}

//==============================================================================
//[MServo::ServoReadMicroseconds] - Get the current pulse width for the specified
//        servo it takes the offsets and reverse into account
//       
//==============================================================================
word MServo::ServoReadMicroseconds(byte bServo)
{
    if (bServo >= NUM_GROUP_MOVE_SERVOS)
        return 0xffff;

    if (_asrvi[bServo].fReverse)
        return 3000 - (_asrvi[bServo].servo.readMicroseconds() - _asrvi[bServo].sPWOffset);
    
    return _asrvi[bServo].servo.readMicroseconds() - _asrvi[bServo].sPWOffset;
}




//==============================================================================
//[ MServo::ServoMoveDegrees1] - Same as Microseconds, but can simply pass in
//        desired angle in 1/10th of a degree and this will do the conversion
//        to microseconds... May combine with above or have inline function...
//==============================================================================
// currently inline function in header file...
//boolean MServo::ServoMoveDegrees1(byte bServo, short sDegrees1)    // Pulse width in tenths of a Degree.
//{
//}

//==============================================================================
//[ MServo::MoveOneServoMicroseconds] - This currently is a wrapper to call
// start, set, and commit, maybe cleanup later to speed things up.
//==============================================================================
boolean  MServo::MoveOneServoMicroseconds(byte bServo, word wPW, word wMoveTime)
{
    GroupMoveStart();                                           // Start a group move
    boolean fRet = ServoMoveMicroseconds(bServo, wPW);          // Desired pulse width in microseconds
    GroupMoveCommit(wMoveTime);                            // how long in milliseconds the move should take
    return fRet;
}



//==============================================================================
//[MServo::WhichGroupMoveServosActive] - Returns a bit mask to let you know
//       which servos are active...  Was made inline in header file
//==============================================================================
//word MServo::WhichGroupMoveServosActive(void);    // returns bit mask for which servos are active.


//==============================================================================
//[MServo::WaitForGPToCompete] - function to call to wait for a group move to
//        complete: may later add in timeout...
//==============================================================================
void MServo::WaitForGMToComplete(unsigned uSGMMask)
{
    // Could put inline...
    while ((_uWhichServosActive & uSGMMask) != 0) {
	delay(1);
    }
}

//==============================================================================
//[MServo::SetServoOffset] - Sets the servo offset for a specific servo
//	 why some moves were not taking as long as expected.
//==============================================================================
boolean MServo::SetServoOffset(byte bServo, short sOffset)
{
    if (bServo >= NUM_GROUP_MOVE_SERVOS)
        return false;
    _asrvi[bServo].sPWOffset = sOffset;
    return true;
}    

//==============================================================================
//[MServo::GetServoOffset] - Retrieve the current offset that is set for a
//        specific servo
//==============================================================================
short MServo::GetServoOffset(byte bServo)
{
    if (bServo >= NUM_GROUP_MOVE_SERVOS)
        return false;
    return  _asrvi[bServo].sPWOffset;
}    

//==============================================================================
//[MServo::DumpTimings] - This is a debug function that I used to help find out 
//	 why some moves were not taking as long as expected.
//==============================================================================
void MServo::DumpTimings(void)
{
#ifdef DEBUG_TIMING
    unsigned i;
    unsigned long tc;
    byte szTemp[80];
    int ich;

    if (g_fDebugOutput) {
        ich = sprintf((char*)szTemp, "\nDump Servo GP Timings\n");
        Serial.write(szTemp, ich);
    
        for (i= 0; i < g_ctcDebug-1; i++) {
    	tc = g_atcDebug[i+1]-g_atcDebug[i];
    
    	if (tc > (unsigned long)0xffff)
        	    ich = sprintf((char*)szTemp, " XXXX");
    	else
    	    ich = sprintf((char*)szTemp, "  %u", (short)tc);
            Serial.write(szTemp, ich);
    	
    	if ((i & 0x7) == 0x7)
                Serial.println("");
        }
        Serial.println("");
    }
    g_ctcDebug = 0;
#endif
}


//==============================================================================
//[MServo::EnableTimer] - Internal function used by the commit function to start
//        up the timer when it finds that there is a group move that is about
//        to become active.
//==============================================================================
void MServo::EnableTimer(void)
{
    // Turn on interrupts
#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega328P__) || defined(__AVR_ATmega1280__)
    TIMSK2 |= 1 << OCIE2A;
#else
    TIMSK |= 1 << OCIE2;
#endif

}



//==============================================================================
//[MServo::DisableTimer] - Internal function called by timer function to
//        disable itself when it finds that there are no more group move 
//        items active.
//==============================================================================
void MServo::DisableTimer(void)
{
     // Turn off interrupts
#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega328P__) || defined(__AVR_ATmega1280__)
    TIMSK2 &= ~(1 << OCIE2A);
#else
    TIMSK &= ~(1 << OCIE2);
#endif

}
